package com.fujitsu.cfmg.sample.source.oui;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.io.StringWriter;
import java.io.Writer;
import java.nio.file.Files;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.Properties;
import java.util.TimeZone;

import javax.xml.namespace.QName;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.xpath.XPath;
import javax.xml.xpath.XPathConstants;
import javax.xml.xpath.XPathExpressionException;
import javax.xml.xpath.XPathFactory;

import org.w3c.dom.Document;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import com.fujitsu.cfmg.sample.source.oui.bean.Component;
import com.fujitsu.cfmg.sample.source.oui.bean.Patch;
import com.fujitsu.cfmg.sample.source.oui.bean.Software;

public class Discover {

	//* A[LeN` */
	private static Map<String, String> architechtureMap = new HashMap<String, String>() {
		{
			//* Linux x86 */
			put("46",  "x86");
			//* Linux x86-64 */
			put("226", "x64");
			//* Microsoft Windows (32-bit) */
			put("912", "x86");
			//* Microsoft Windows x64 (64-bit) */
			put("233", "x64");
			//* Sun Solaris SPARC (32-bit) */
			put("453", "sparc-32");
			//* Sun Solaris SPARC (64-bit) */
			put("23",  "sparc-64");
		}
	};
	//* o͐fBNg */
	private static String outputDir;
	//* Iy[eBOVXe */
	private static String osName;
	//* CXg[σ\tgEFA */
	private static List<Software> softwareList = new ArrayList<Software>();
	//* Kpσpb` */
	private static List<Patch> patchList = new ArrayList<Patch>();
	
	/**
	 *  Oralce Universal Installer\tgEFA񂨂уpb`W܂B
	 *  ܂AW\tgEFACSVt@Cƃpb`CSVt@Cɏo͂܂B
	 *  
	 *  @param args[0] \tgEFACSVt@Cƃpb`CSVt@C̏o͐fBNg
	 */
	public static void main(String args[]) {

		outputDir = args[0];
		
		osName = System.getProperty("os.name");

		if (osName == null) {
			System.err.println("ERROR99999 A system error has occurred. [Details:The system property(os.name) not found.]");
			System.exit(1);
		}
		
		/** 1.a CxgfBNg̎擾 */
		List<String> centralInventoryDirList = new ArrayList<String>();
		
		if(osName.indexOf("Windows") >= 0) {
			List<CommandResult> resultList = new ArrayList<CommandResult>();

			CommandResult result64 = executeCommand(Arrays.asList("REG", "QUERY", "HKLM\\SOFTWARE\\ORACLE", "/v", "inst_loc"));
			if (result64.getRet() == 0) {
				resultList.add(result64);
			}
			CommandResult result32 = executeCommand(Arrays.asList("REG", "QUERY", "HKLM\\SOFTWARE\\Wow6432Node\\ORACLE", "/v", "inst_loc"));
			if (result32.getRet() == 0) {
				resultList.add(result32);
			}
			
			if (resultList.size() == 0) {
				System.out.println("INFO00001 The registry key not found. [Key:HKLM\\SOFTWARE\\ORACLE\\inst_loc]");
				outputCsvFile();
				System.exit(0);
			}
			
			for (CommandResult result : resultList) {
				for (String line : result.getStdout().split(System.getProperty("line.separator"))) {
					line = line.trim();
					
					if (line.startsWith("inst_loc")) {
						line = line.substring("inst_loc".length()).trim();
						
						if (line.startsWith("REG_SZ")) {
							centralInventoryDirList.add(line.substring("REG_SZ".length()).trim());
							break;
						}
					}
				}
			}

		} else if (osName.indexOf("Linux") >= 0 || osName.indexOf("SunOS") >= 0) {
			String centralInventoryPointerFile = null;
			if (osName.indexOf("Linux") >= 0) centralInventoryPointerFile = "/etc/oraInst.loc";
			if (osName.indexOf("SunOS") >= 0) centralInventoryPointerFile = "/var/opt/oracle/oraInst.loc";
			
			if (!new File(centralInventoryPointerFile).exists()) {
				System.out.println("INFO00002 The central inventory pointer file not found. [File path:" + centralInventoryPointerFile + "]");
				outputCsvFile();
				System.exit(0);
			}
			
			String centralInventoryDir = null;
			FileInputStream fileInputStream = null;
			try {
				Properties properties = new Properties();
				fileInputStream = new FileInputStream(centralInventoryPointerFile);
				properties.load(fileInputStream);
				centralInventoryDir = properties.getProperty("inventory_loc");

			} catch (Throwable e) {
				System.err.println("ERROR99999 A system error has occurred. [Details:Failed to read the file(" + centralInventoryPointerFile + ").]");
				e.printStackTrace();
				System.exit(1);

			} finally {
				try {
					fileInputStream.close();
				} catch (IOException e) {
					e.printStackTrace();
				}
			}

			if (centralInventoryDir != null) {
				centralInventoryDirList.add(centralInventoryDir);
			} else if (centralInventoryDir == null) {
				System.err.println("ERROR99999 A system error has occurred. [Details:The property key(inventory_loc) not found.]");
				System.exit(1);
			}

		} else {
			System.err.println("ERROR99999 A system error has occurred. [Details:The system property os.name(" + osName + ") is invalid.]");
			System.exit(1);
		}

		/** 1.b Oraclez[̎擾 */
		List<String> oracleHomeList = new ArrayList<String>();
		for (String centralInventoryDir : centralInventoryDirList) {
			String centralInventoryFilePath = centralInventoryDir + File.separator + "ContentsXML" + File.separator + "inventory.xml";

			if (!new File(centralInventoryFilePath).exists()) {
				System.err.println("ERROR99999 A system error has occurred. [Details:The central inventory file(" + centralInventoryFilePath + ") not found.]");
				System.exit(1);
			}

			Document centralInventoryDoc = null;
			try {
				DocumentBuilder builder = DocumentBuilderFactory.newInstance().newDocumentBuilder();
				centralInventoryDoc = builder.parse(new File(centralInventoryFilePath));
			} catch (Throwable e) {
				System.err.println("ERROR99999 A system error has occurred. [Details:Failed to read the XML file(" + centralInventoryFilePath + ").]");
				e.printStackTrace();
				System.exit(1);
			}
			
			NodeList oracleHomeNodeList = (NodeList) evaluate("/INVENTORY/HOME_LIST/HOME[@LOC][@TYPE='O']", centralInventoryDoc, XPathConstants.NODESET);

			for (int i = 0; i < oracleHomeNodeList.getLength(); i++ ) {
				Node oracleHomeNode = oracleHomeNodeList.item(i);

				NamedNodeMap atts = oracleHomeNode.getAttributes();
				Node attNode = atts.getNamedItem("REMOVED");
				if (attNode != null && "T".equals(attNode.getNodeValue())) {
					continue;
				}

				String oracleHome = atts.getNamedItem("LOC").getNodeValue();
				if (oracleHome != null && !oracleHome.isEmpty()) {
					if (!oracleHomeList.contains(oracleHome)) {
						oracleHomeList.add(oracleHome);
					}
				}
			}
		}

		if (oracleHomeList.size() == 0) {
			System.out.println("INFO00003 Oracle home not found.");
			outputCsvFile();
			System.exit(0);
		}

		for (String oracleHome : oracleHomeList) {
			/** 2.a Oracle Database^Oracle Fusion Middleware̔ */
			String registryFilePath = oracleHome + File.separator + "inventory" + File.separator + "registry.xml";

			if (new File(registryFilePath).exists()) {
				/** 2.b Oracle Fusion Middleware */
				discoverOracleFusionMiddleware(oracleHome);

			} else {
				/** 2.c Oracle Database */
				discoverOracleDatabase(oracleHome);
			}
		}

		/** 3. \tgEFACSVt@Cƃpb`CSVt@C̏o */
		outputCsvFile();
	}

	/**
	 *  w肳ꂽOraclez[Oracle Fusion Middleware
	 *  Oracle Fusion MiddlewareɓKpꂽpb`W܂B
	 *  
	 *  @param oracleHome Oraclez[
	 */
	private static void discoverOracleFusionMiddleware(String oracleHome) {
		
		/** 2.b.1 viewInventoryR}h̎s */
		String viewInventoryFileName;
		if (osName.indexOf("Windows") >= 0) {
			viewInventoryFileName = "viewInventory.cmd";
		} else {
			viewInventoryFileName = "viewInventory.sh";
		}
		String viewInventoryFilePath = oracleHome + File.separator + "oui" + File.separator + "bin" + File.separator + viewInventoryFileName;
		
		if (!new File(viewInventoryFilePath).exists()) {
			System.err.println("ERROR99999 A system error has occurred. [Details:viewInventory(" + viewInventoryFilePath + ") not found.]");
			System.exit(1);
		}
		
		String outputFilePath = System.getProperty("user.dir") + File.separator + new Date().getTime() + ".xml";
		List<String> commandLine = Arrays.asList(viewInventoryFilePath, "-oracle_home", oracleHome, "-output_format", "xml", "-output_file", outputFilePath);
		
		CommandResult result = executeCommand(commandLine);
		if (result.getRet() != 0) {
			System.err.println("ERROR00001 Failed to execute the command. [Command line:" + commandLine.toString() + "]");
			System.err.println("--------------------------------");
			System.err.println(result.getStdout());
			System.err.println("--------------------------------");
			System.err.println(result.getRet());
			System.exit(1);
		}
		if (!new File(outputFilePath).exists()) {
			System.err.println("ERROR00002 The output file not found. [Command line:" + commandLine.toString() + "]");
			System.err.println("--------------------------------");
			System.err.println(result.getStdout());
			System.err.println("--------------------------------");
			System.err.println(result.getRet());
			System.exit(1);
		}
		
		/** 2.b.2 \tgEFAƃpb`̎擾 */
		Document inventoryDoc = null;
		try {
			DocumentBuilder builder = DocumentBuilderFactory.newInstance().newDocumentBuilder();
			inventoryDoc = builder.parse(new File(outputFilePath));
		} catch (Throwable e) {
			System.err.println("ERROR99999 A system error has occurred. [Details:Failed to read the XML file(" + outputFilePath + ").]");
			e.printStackTrace();
			System.exit(1);

		} finally {
			new File(outputFilePath).delete();
		}

		List<Software> distributionList = new ArrayList<Software>();
		List<Patch> distributionPatchList = new ArrayList<Patch>();

		double distributionCount = (double) evaluate("count(/oracle-home/distributions/distribution)", inventoryDoc, XPathConstants.NUMBER);
		for (int i = 1; i <= distributionCount; i++ ) {
			Software distribution = new Software();

			String softwareName = (String) evaluate("/oracle-home/distributions/distribution[" + i + "]/@name", inventoryDoc, XPathConstants.STRING);
			distribution.setSoftwareName(softwareName);
			
			String version = (String) evaluate("/oracle-home/distributions/distribution[" + i + "]" + "/@version", inventoryDoc, XPathConstants.STRING);
			distribution.setVersion(version);

			distribution.setLocalId(softwareList.size() + distributionList.size() + 1);
			distribution.setInstallPath(oracleHome);
			distribution.setVendor("Oracle Corporation");
			distributionList.add(distribution);

			NodeList patchIdNodeList = (NodeList) evaluate("/oracle-home/distributions/distribution[" + i + "]/feature-sets/feature-set/components/component/patches/patch/@patch-id", inventoryDoc, XPathConstants.NODESET);
			List<String> patchIdList = new ArrayList<String>();
			for (int j = 0; j < patchIdNodeList.getLength(); j++ ) {
				String patchId = patchIdNodeList.item(j).getNodeValue();

				if (patchId != null && !patchId.isEmpty() && !patchIdList.contains(patchId)) {
					patchIdList.add(patchId);
					Patch distributionPatch = new Patch();

					distributionPatch.setLocalId(patchList.size() + distributionPatchList.size() + 1);
					distributionPatch.setPatchName(patchId);
					distributionPatch.setSoftware(distribution.getLocalId());
					distributionPatch.setSoftwareName(distribution.getSoftwareName() + " " + distribution.getVersion());
					distributionPatchList.add(distributionPatch);
				}

			}
		}

		/** 2.b.3 opatchR}h̎s */
		/** 2.b.4 pb`̏ڍ׏̎擾 */
		Map<String, Patch> patchDetailMap = getPatchDetailInfo(oracleHome);
		for (Patch distributionPatch : distributionPatchList) {
			String patchName = distributionPatch.getPatchName();
			if (patchDetailMap.containsKey(patchName)) {
				distributionPatch.setApplyDate(patchDetailMap.get(patchName).getApplyDate());
				distributionPatch.setTitle(patchDetailMap.get(patchName).getTitle());
			}
		}

		softwareList.addAll(distributionList);
		patchList.addAll(distributionPatchList);
	}

	/**
	 *  w肳ꂽOraclez[Oracle Database
	 *  Oracle DatabaseɓKpꂽpb`W܂B
	 *  
	 *  @param oracleHome Oraclez[
	 */
	private static void discoverOracleDatabase(String oracleHome) {
		/** 2.c.1 \tgEFAƃpb`̎擾 */
		String localInventoryFilePath = oracleHome + File.separator + "inventory" + File.separator + "ContentsXML" + File.separator + "comps.xml";

		if (!new File(localInventoryFilePath).exists()) {
			System.out.println("INFO00004 The local inventory file not found. [Oracle home:" + oracleHome + "][File path:" + localInventoryFilePath + "]");
			return;
		}
		
		Document localInventoryDoc = null;
		try {
			DocumentBuilder builder = DocumentBuilderFactory.newInstance().newDocumentBuilder();
			localInventoryDoc = builder.parse(new File(localInventoryFilePath));
		} catch (Throwable e) {
			System.err.println("ERROR99999 A system error has occurred. [Details:Failed to read the XML file(" + localInventoryFilePath + ").]");
			e.printStackTrace();
			System.exit(1);
		}

		List<Software> topLevelCompList = new ArrayList<Software>();
		double topLevelCompCount = (double) evaluate("count(/PRD_LIST/TL_LIST/COMP)", localInventoryDoc, XPathConstants.NUMBER);
		for (int i = 1; i <= topLevelCompCount; i++ ) {
			Software topLevelComp = new Software();

			String softwareName = (String) evaluate("/PRD_LIST/TL_LIST/COMP[" + i + "]/EXT_NAME/text()", localInventoryDoc, XPathConstants.STRING);
			topLevelComp.setSoftwareName(softwareName);

			String version = (String) evaluate("/PRD_LIST/TL_LIST/COMP[" + i + "]/@VER", localInventoryDoc, XPathConstants.STRING);
			topLevelComp.setVersion(version);

			String installDate = (String) evaluate("/PRD_LIST/TL_LIST/COMP[" + i + "]/@INSTALL_TIME", localInventoryDoc, XPathConstants.STRING);
			SimpleDateFormat format = new SimpleDateFormat("yyyy.MMM.dd HH:mm:ss z", Locale.US);
			try {
				Date date = format.parse(installDate);
				format = new SimpleDateFormat("yyyy-MM-dd HH:mm:ss");
				format.setTimeZone(TimeZone.getDefault());
				installDate = format.format(date);
				topLevelComp.setInstallDate(installDate);
			} catch (ParseException e) {
				System.err.println("WARN00002 Invalid value. [File path:" + localInventoryFilePath + "][XPath:" + "/PRD_LIST/TL_LIST/COMP[" + i + "]/@INSTALL_TIME" + "]");
				e.printStackTrace();
			}

			String architecture = (String) evaluate("/PRD_LIST/TL_LIST/COMP[" + i + "]/PLAT_IDX_MAP/PLAT/@LIST", localInventoryDoc, XPathConstants.STRING);
			if (architechtureMap.containsKey(architecture)) {
				topLevelComp.setArchitecture(architechtureMap.get(architecture));
			} else {
				System.err.println("WARN00002 Invalid value. [File path:" + localInventoryFilePath + "][XPath:" + "/PRD_LIST/TL_LIST/COMP[" + i + "]/PLAT_IDX_MAP/PLAT/@LIST" + "]");
			}

			String edition = (String) evaluate("/PRD_LIST/TL_LIST/COMP[" + i + "]/INST_TYPE_LIST/INST_TYPE/@NAME", localInventoryDoc, XPathConstants.STRING);
			topLevelComp.setEdition(edition);

			double compCount = (double) evaluate("count(/PRD_LIST/TL_LIST/COMP[" + i + "]/DEP_LIST/DEP)", localInventoryDoc, XPathConstants.NUMBER);
			List<Component> components = new ArrayList<Component>();
			for (int j = 1; j <= compCount; j++) {
				Component component = new Component();

				String compName = (String) evaluate("/PRD_LIST/TL_LIST/COMP[" + i + "]/DEP_LIST/DEP[" + j + "]/@NAME", localInventoryDoc, XPathConstants.STRING);
				component.setName(compName);

				String compVersion = (String) evaluate("/PRD_LIST/TL_LIST/COMP[" + i + "]/DEP_LIST/DEP[" + j + "]/@VER", localInventoryDoc, XPathConstants.STRING);
				component.setVersion(compVersion);

				components.add(component);
			}

			topLevelComp.setComponents(components);
			topLevelComp.setLocalId(softwareList.size() + topLevelCompList.size() + 1);
			topLevelComp.setInstallPath(oracleHome);
			topLevelComp.setVendor("Oracle Corporation");
			topLevelCompList.add(topLevelComp);
		}

		List<Patch> oneOffList = new ArrayList<Patch>();
		double oneoffCount = (double) evaluate("count(/PRD_LIST/ONEOFF_LIST/ONEOFF)", localInventoryDoc, XPathConstants.NUMBER);
		for (int i = 1; i <= oneoffCount; i++ ) {

			Patch oneOff = new Patch();

			String patchName = (String) evaluate("/PRD_LIST/ONEOFF_LIST/ONEOFF[" + i + "]" + "/@REF_ID", localInventoryDoc, XPathConstants.STRING);
			oneOff.setPatchName(patchName);

			double compCount = (double) evaluate("count(" + "/PRD_LIST/ONEOFF_LIST/ONEOFF[" + i + "]" + "/REF_LIST/REF)", localInventoryDoc, XPathConstants.NUMBER);
			List<Component> components = new ArrayList<Component>();
			for (int j = 1; j <= compCount; j++) {
				Component component = new Component();

				String compName = (String) evaluate("/PRD_LIST/ONEOFF_LIST/ONEOFF[" + i + "]" + "/REF_LIST/REF[" + j + "]/@NAME", localInventoryDoc, XPathConstants.STRING);
				component.setName(compName);

				String compVersion = (String) evaluate("/PRD_LIST/ONEOFF_LIST/ONEOFF[" + i + "]" + "/REF_LIST/REF[" + j + "]/@VER", localInventoryDoc, XPathConstants.STRING);
				component.setVersion(compVersion);

				components.add(component);
			}

			oneOff.setComponents(components);

			/** 2.c.2 \tgEFAƃpb`̕Rt */
			for (Software topLevelComp : topLevelCompList) {
				for (Component oneOffComp : oneOff.getComponents()) {
					boolean applied = false;
					for (Component comp : topLevelComp.getComponents()) {
						if (oneOffComp.getName().equals(comp.getName()) && oneOffComp.getVersion().equals(comp.getVersion())) {
							Patch patch = new Patch();
							patch.setLocalId(patchList.size() + oneOffList.size() + 1);
							patch.setPatchName(oneOff.getPatchName());
							patch.setSoftware(topLevelComp.getLocalId());
							patch.setSoftwareName(topLevelComp.getSoftwareName() + " " + topLevelComp.getVersion());
							oneOffList.add(patch);
							applied = true;
							break;
						}
					}
					if (applied) break;
				}
			}
		}

		/** 2.c.3 opatchR}h̎s */
		/** 2.c.4 pb`̏ڍ׏̎擾 */
		Map<String, Patch> patchDetailMap = getPatchDetailInfo(oracleHome);
		for (Patch oneOff : oneOffList) {
			String patchName = oneOff.getPatchName();
			if (patchDetailMap.containsKey(patchName)) {
				oneOff.setApplyDate(patchDetailMap.get(patchName).getApplyDate());
				oneOff.setTitle(patchDetailMap.get(patchName).getTitle());
			}
		}

		softwareList.addAll(topLevelCompList);
		patchList.addAll(oneOffList);
	}

	/**
	 *  OPatchopatchR}hsāA
	 *  Oraclez[ɓKpꂽpb`̏ڍ׏W܂B
	 *  
	 * @param oracleHome Oraclez[
	 * @param pb`̏ڍ׏
	 */
	private static Map<String, Patch> getPatchDetailInfo(String oracleHome) {
		Map<String, Patch> patchMap = new HashMap<String, Patch>();

		/** opatchR}h̎s */
		String opatchFileName;
		if (osName.indexOf("Windows") >= 0) {
			opatchFileName = "opatch.bat";
		} else {
			opatchFileName = "opatch";
		}
		String opatchFilePath = oracleHome + File.separator + "OPatch" + File.separator + opatchFileName;
		if (!new File(opatchFilePath).exists()) {
			System.err.println("WARN00001 The opatch command not found. [Oracle home:" + oracleHome + "][File path:" + opatchFilePath + "]");
			return patchMap;
		}

		String outputFilePath = "";
		List<String> commandLine = new ArrayList<String>();
		if (osName.indexOf("Windows") >= 0) {
			outputFilePath = "." + File.separator + new Date().getTime() + ".xml";
			commandLine = Arrays.asList(opatchFilePath, "lsinventory", "-oh", oracleHome, "-xml", outputFilePath);

		} else {
			String oracleUser = "";
			try {
				oracleUser = Files.getOwner(new File(oracleHome).toPath()).getName();
			} catch (IOException e) {
				System.err.println("ERROR99999 A system error has occurred. [Details:Failed to get the owner of the directory(" + oracleHome + ").]");
				e.printStackTrace();
				System.exit(1);
			}
			outputFilePath = System.getProperty("java.io.tmpdir") + File.separator + new Date().getTime() + ".xml";
			commandLine = Arrays.asList("su", "-", oracleUser, "-c", opatchFilePath + " lsinventory -oh " + oracleHome + " -xml " + outputFilePath);
		}

		CommandResult result = executeCommand(commandLine);
		if (result.getRet() != 0) {
			System.err.println("ERROR00001 Failed to execute the command. [Command line:" + commandLine.toString() + "]");
			System.err.println("--------------------------------");
			System.err.println(result.getStdout());
			System.err.println("--------------------------------");
			System.err.println(result.getRet());
			System.exit(1);
		}
		if (!new File(outputFilePath).exists()) {
			System.err.println("ERROR00002 The output file not found. [Command line:" + commandLine.toString() + "]");
			System.err.println("--------------------------------");
			System.err.println(result.getStdout());
			System.err.println("--------------------------------");
			System.err.println(result.getRet());
			System.exit(1);
		}

		/** pb`̏ڍ׏̎擾 */
		Document inventoryDoc = null;
		try {
			DocumentBuilder builder = DocumentBuilderFactory.newInstance().newDocumentBuilder();
			inventoryDoc = builder.parse(new File(outputFilePath));
		} catch (Throwable e) {
			System.err.println("ERROR99999 A system error has occurred. [Details:Failed to read the XML file(" + outputFilePath + ").]");
			e.printStackTrace();
			System.exit(1);

		} finally {
			new File(outputFilePath).delete();
		}

		double patchCount = (double) evaluate("count(/InventoryInstance/patches/patch)", inventoryDoc, XPathConstants.NUMBER);
		for (int i = 1; i <= patchCount; i++ ) {
			Patch patch = new Patch();

			String patchName = (String) evaluate("/InventoryInstance/patches/patch[" + i + "]/patchID/text()", inventoryDoc, XPathConstants.STRING);
			patch.setPatchName(patchName);

			String applyDate = (String) evaluate("/InventoryInstance/patches/patch[" + i + "]/appliedDate/text()", inventoryDoc, XPathConstants.STRING);
			Date date = null;
			SimpleDateFormat format = new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ss.SSSXXX");
			try {
				date = format.parse(applyDate);
			} catch (ParseException e) {
				try {
					format = new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ssXXX");
					date = format.parse(applyDate);
				} catch (ParseException e1) {
					System.err.println("WARN00002 Invalid value. [File path:" + outputFilePath + "][XPath:" + "/InventoryInstance/patches/patch[" + i + "]/appliedDate/text()" + "]");
					e.printStackTrace();
				}
			}
			if(date != null) {
				format = new SimpleDateFormat("yyyy-MM-dd HH:mm:ss");
				format.setTimeZone(TimeZone.getDefault());
				patch.setApplyDate(format.format(date));
			}

			String title = (String) evaluate("/InventoryInstance/patches/patch[" + i + "]/patchDescription/text()", inventoryDoc, XPathConstants.STRING);
			patch.setTitle(title);

			patchMap.put(patch.getPatchName(), patch);
		}
		return patchMap;
	}

	/**
	 *  w肳ꂽCSVtB[hGXP[v܂B
	 *  
	 * @param value CSVtB[h
	 * @return GXP[vꂽtB[h
	 */
	private static String escape(String value) {
		if (value == null) value = "";
		if (value.indexOf(",") >= 0 || value.indexOf("\"") >= 0 || value.indexOf("\r") >= 0 || value.indexOf("\n") >= 0 || value.indexOf(",") >= 0) {
			StringBuffer buffer = new StringBuffer();
			if (value.indexOf(",") >= 0 || value.indexOf("\"") >= 0) buffer.append("\"");
			char[] charArray = value.toCharArray();
			for (int i = 0; i < charArray.length; ++i) {
				int c = charArray[i];
				if (c == '\r' ) continue;
				if (c == '\n' ) continue;
				buffer.append(c);
				if (c == '\"') {
					buffer.append(c);
				}
			}
			if (value.indexOf(",") >= 0 || value.indexOf("\"") >= 0) buffer.append("\"");
			value = value.toString();
		}
		return value;
	}

	/**
	 *  \tgEFAƃpb`
	 *  \tgEFACSVt@Cƃpb`CSVt@Cɏo͂܂B
	 *  
	 */
	private static void outputCsvFile() {
		String softwareCsvFilePath = outputDir + File.separator + "softwareinfo.csv";
		String patchCsvFilePath = outputDir + File.separator + "patchinfo.csv";

		StringWriter writer = new StringWriter();
		for (Software software : softwareList) {
			writer.write(escape(String.valueOf(software.getLocalId())));
			writer.write(",");
			writer.write(escape(software.getSoftwareName()));
			writer.write(",");
			writer.write(escape(software.getVersion()));
			writer.write(",");
			writer.write(escape(software.getEdition()));
			writer.write(",");
			writer.write(escape(software.getArchitecture()));
			writer.write(",");
			writer.write(escape(software.getOsType()));
			writer.write(",");
			writer.write(escape(software.getVendor()));
			writer.write(",");
			writer.write(escape(software.getInstallDate()));
			writer.write(",");
			writer.write(escape(software.getInstallPath()));
			writer.write(System.getProperty("line.separator"));
		}
		writeFile(softwareCsvFilePath, writer);

		if (patchList.size() > 0) {
			writer = new StringWriter();
			for (Patch patch : patchList) {
				writer.write(escape(String.valueOf(patch.getLocalId())));
				writer.write(",");
				writer.write(escape(patch.getPatchName()));
				writer.write(",");
				writer.write(escape(String.valueOf(patch.getSoftware())));
				writer.write(",");
				writer.write(escape(patch.getSoftwareName()));
				writer.write(",");
				writer.write(escape(patch.getOsType()));
				writer.write(",");
				writer.write(escape(patch.getApplyDate()));
				writer.write(",");
				writer.write(escape(patch.getTitle()));
				writer.write(System.getProperty("line.separator"));
			}
			writeFile(patchCsvFilePath, writer);
		}
	}

	/**
	 *  w肳ꂽt@Cɕ݂܂B
	 *  
	 * @param filePath t@CpX
	 * @param writer Xg[C^[
	 */
	private static void writeFile(String filePath, Writer writer) {
		FileOutputStream fos = null;
		BufferedWriter bw = null;
		try {
			fos = new FileOutputStream(new File(filePath));
			bw = new BufferedWriter(new OutputStreamWriter(fos));
			bw.write(writer.toString());
			bw.flush();
		} catch (Throwable e) {
			System.err.println("ERROR99999 A system error has occurred. [Details:Failed to write the file(" + filePath + ").]");
			e.printStackTrace();
			System.exit(1);
		} finally {
			if (writer != null) {
				try {
					writer.close();
				} catch (Throwable e) {
					e.printStackTrace();
				}
			}
			if (fos != null) {
				try {
					fos.close();
				} catch (Throwable e) {
					e.printStackTrace();
				}
			}
			if (bw != null) {
				try {
					bw.close();
				} catch (Throwable e) {
					e.printStackTrace();
				}
			}
		}
	}

	/**
	 *  R}hs܂B
	 *  
	 * @param commandLine sR}h
	 * @return R}h̎s
	 */
	private static CommandResult executeCommand(List<String> commandLine) {
		CommandResult result = new CommandResult();

		ProcessBuilder processBuilder = new ProcessBuilder(commandLine);
		processBuilder.redirectErrorStream(true);
		try {
			final Process process = processBuilder.start();
			final StringBuffer buffer = new StringBuffer();
			final Thread thread = new Thread() {
				public void run() {
					BufferedReader bufferedReader = null;
					try {
						bufferedReader = new BufferedReader(new InputStreamReader(process.getInputStream()));
						String line = null;
						while ((line = bufferedReader.readLine()) != null) {
							buffer.append(line);
							buffer.append(System.getProperty("line.separator"));
						}
					} catch (Throwable e) {
						e.printStackTrace();
					} finally {
						try {
							bufferedReader.close();
						} catch (IOException e) {
							e.printStackTrace();
						}
					}
				}
			};
			thread.start();

			result.setRet(process.waitFor());
			result.setStdout(buffer.toString());
		} catch (Throwable e) {
			System.err.println("ERROR99999 A system error has occurred. [Details:Failed to execute the command" + commandLine.toString() + ".");
			e.printStackTrace();
			System.exit(1);
		}

		return result;
	}

	/**
	 *  w肳ꂽhLgXpathŕ]܂B
	 *  
	 * @param expression Xpath
	 * @param document hLg
	 * @param qName ߂ľ^
	 * @return ̕]̌
	 */
	private static Object evaluate(String expression, Document document, QName qName) {
		XPathFactory factory = XPathFactory.newInstance();
		XPath xpath = factory.newXPath();
		Object object = null;
		try {
			object = xpath.evaluate(expression, document, qName);
		} catch (XPathExpressionException e) {
			System.err.println("ERROR99999 A system error has occurred. [Details:The expression(" + expression + ") is invalid.]");
			e.printStackTrace();
			System.exit(1);
		}
		return object;
	}
}